'''
Created on 2021-04-29

@author: lucianovilas@gmail.com

'''


import numpy as np
from .utils import make_clusters, check_clusters, check_labels


# Auxiliar constants and functions

ND_Epsilon=1e-07 # A very small value to avoid division by zero


def f1(p,r):
    """ Harminic mean """
    return 2 * p * r / (p + r + ND_Epsilon)



### Api functions    

        
# K-Metric only calculate
def k_metric(T,P):
    """Compute: K-Metric
   
    Parameters
    ----------
    :param T: clusters containing the ground truth cluster labels.
    :param P: clusters containing the predicted cluster labels.
    
    Returns 
    -------
    :return float AAP: KMetric recall
    :return float ACP: KMetric precision
    :return float K: KMetric geometric mean
    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).    
    """
    
    pIndex, cSize = {}, {}
    instSum, aapSum, acpSum = 0,0,0
    for i,Pi in enumerate(P):
        for p in P[Pi]:
            pIndex[p]=i
        cSize[i]=len(P[Pi])
        
    instSum=0
    for Tj in T:
        len_Tj = len(T[Tj])
        instSum += len_Tj
        
        tMap = {}
        for t in T[Tj]:
            if pIndex[t] not in tMap:
                tMap[pIndex[t]] = 0
            tMap[pIndex[t]] += 1
        for k,v in tMap.items():
            aapSum += v**2 / len_Tj
            acpSum += v**2 / cSize[k]

    AAP = aapSum/instSum
    ACP = acpSum/instSum
    
    return ACP, AAP, np.sqrt(AAP*ACP)


# All-In-One metrics calculate
def all_metrics(T,P):
    """Compute all metrics for name disambiguation methods.
    Compute: ClusterF-Metric, K-Metric, B^3-Metric, SE&LE-Metric and PairwiseF-Metric
   
    Parameters
    ----------
    :param T: map of clusters containing the ground truth cluster labels.
    :param P: map of clusters containing the predicted cluster labels.
    
    Returns 
    -------
    :return float cR: ClusterFMetric recall
    :return float cP: ClusterFMetric precision
    :return float f_score: ClusterFMetric f_score
    
    :return float AAP: KMetric recall
    :return float ACP: KMetric precision
    :return float K: KMetric geometric mean
    
    :return float bR: BCubedMetric recall
    :return float bP: BCubedMetric precision
    :return float f_score: BCubedMetric f_score

    :return float SE: SELEMetric recall
    :return float LE: SELEMetric precision
    :return float f_score: SELEMetric f_score

    :return float pR: PairwiseFMetric recall
    :return float pP: PairwiseFMetric precision
    :return float f_score: PairwiseFMetric f_score

    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).    
    """
    
    pIndex = {} # Common to all measures
    cSize = {}  # ClusterF-Metric, K-Metric, B^3-Metric and SE&LE-Metric
    cMatch = 0  # ClusterF-Metric
    instSum, aapSum, acpSum = 0,0,0 #  K-Metric and B^3-Metric
    spSum, lmSum, instTrSum, instPrSum = 0,0,0,0 # SE&LE-Metric
    pairPrSum, pairTrSum, pairInstSum = 0,0,0 # PairwiseF-Metric
    
    for i,Pi in enumerate(P):
        for p in P[Pi]:
            pIndex[p]=i
        cSize[i]=len(P[Pi])
        pairPrSum += len(P[Pi])*(len(P[Pi])-1)/2.0
        
    instSum=0
    for Tj in T:
        len_Tj = len(T[Tj])
        instSum += len_Tj
        pairTrSum += len_Tj*(len_Tj-1)/2.0
        
        tMap = {}
        for t in T[Tj]:
            if pIndex[t] not in tMap:
                tMap[pIndex[t]] = 0
            tMap[pIndex[t]] += 1
        
        maxKey, maxValue = 0,0            
        for k,v in tMap.items():
            if v == len_Tj and cSize[k] == len_Tj:
                cMatch += 1
            aapSum += v**2 / len_Tj
            acpSum += v**2 / cSize[k]            
            if v > maxValue:
                maxValue = v
                maxKey = k
            pairInstSum += v*(v-1)/2.0
        
        spSum += len_Tj - maxValue
        lmSum += cSize[maxKey] - maxValue
        instTrSum += len_Tj
        instPrSum += cSize[maxKey]
    
    cR, cP   = cMatch / len(T), cMatch / len(P)
    AAP, ACP =  aapSum/instSum, acpSum/instSum
    bR, bP   = AAP, ACP 
    SE, LE   = spSum/instTrSum, lmSum/instPrSum
    pR, pP   = pairInstSum/pairTrSum, pairInstSum/pairPrSum
    
    return {'ClusterFMetric': (cP, cR, f1(cP,cR)),\
            'KMetric': (ACP, AAP, np.sqrt(ACP*AAP)),\
            'BCubedMetric': (bP, bR, f1(bP,bR)),\
            'SELEMetric': (SE, LE, f1(SE,LE)),\
            'PairwiseFMetric': (pP, pR, f1(pP,pR))}



### Api classes    
    
class ND_MetricBase():
    """
        Base class for initiate and check inputs...
    """
    
    def __init__(self, labels_true, labels_pred):

        self.T, self.P = self.valid_inputs(labels_true, labels_pred)
        self.pIndex = {} # Common to all measures

    def valid_inputs(self, true_inputs, pred_inputs):
        if isinstance(true_inputs, (np.ndarray, list)) and isinstance(pred_inputs, (np.ndarray, list)):
            return make_clusters( *check_labels(true_inputs, pred_inputs) )
        elif isinstance(true_inputs, (dict)) and isinstance(pred_inputs, (dict)):
            return check_clusters(true_inputs, pred_inputs)
        else:     
            raise ValueError('Mismatch formats')
        
    def score(self):
        pass
    
    def __call__(self):
        return self.score()    




class ND_ClusterF(ND_MetricBase):
    """
    Class ClusterF Metric 
    
    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).         

    """
    
    def __init__(self, labels_true, labels_pred):
        super().__init__(labels_true, labels_pred)

    def score(self):
        cSize = {}  # Cluster-F, K-Metric, B^3 and SE&LE
        cMatch = 0  # Cluster-F

        for i,Pi in enumerate(self.P):
            for p in self.P[Pi]:
                self.pIndex[p]=i
            cSize[i]=len(self.P[Pi])

        instSum=0
        for Tj in self.T:
            len_Tj = len(self.T[Tj])
            instSum += len_Tj
            
            tMap = {}
            for t in self.T[Tj]:
                if self.pIndex[t] not in tMap:
                    tMap[self.pIndex[t]] = 0
                tMap[self.pIndex[t]] += 1

            for k,v in tMap.items():
                if v == len_Tj and cSize[k] == len_Tj:
                    cMatch += 1

        cR, cP   = cMatch / len(self.T), cMatch / len(self.P)
        
        return cP, cR, f1(cP, cR)
    
    
        
class ND_KMetric (ND_MetricBase):
    """
    Class KMetric Metric 
    
    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).         

    """
    
    def __init__(self, labels_true, labels_pred):
        super().__init__(labels_true, labels_pred)

    def score(self):
        
        self.cSize = {}
        instSum, aapSum, acpSum = 0,0,0
        
        for i,Pi in enumerate(self.P):
            for p in self.P[Pi]:
                self.pIndex[p]=i
            self.cSize[i]=len(self.P[Pi])

        instSum=0
        for Tj in self.T:
            len_Tj = len(self.T[Tj])
            instSum += len_Tj
            
            tMap = {}
            for t in self.T[Tj]:
                if self.pIndex[t] not in tMap:
                    tMap[self.pIndex[t]] = 0
                tMap[self.pIndex[t]] += 1
            for k,v in tMap.items():
                aapSum += v**2 / len_Tj
                acpSum += v**2 / self.cSize[k]

        AAP = aapSum / instSum
        ACP = acpSum / instSum
        
        return ACP, AAP, np.sqrt(ACP*AAP)
    

class ND_BCubedMetric(ND_KMetric):
    """
    Class BCubedMetric Metric 
    
    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).         

    """   
        
    def __init__(self, labels_true, labels_pred):
        super().__init__(labels_true, labels_pred)
    
    def score(self):
        bP, bR, _ = super().score()
        
        return bP, bR, f1(bP, bR)
    


class ND_PairwiseFMetric (ND_MetricBase):
    """
    Class PairwiseFMetric Metric 
    
    Reference
    ---------
    Kim, J. "A fast and integrative algorithm for 
    clustering performance evaluation in author name disambiguation."
    Scientometrics (2019): 661-681, 120(2).         

    """
    
    def __init__(self, labels_true, labels_pred):
        super().__init__(labels_true, labels_pred)
    
    def score(self):
        
        pairPrSum=0
        pairTrSum=0
        pairInstSum=0  
        
        for i,Pi in enumerate(self.P):
            for p in self.P[Pi]:
                self.pIndex[p]=i
                len_Pi = len(self.P[Pi])
            pairPrSum += len_Pi * (len_Pi-1) / 2.0

        for Tj in self.T:
            len_Tj = len(self.T[Tj])
            pairTrSum += len_Tj * (len_Tj-1) / 2.0
            
            tMap = {}
            for t in self.T[Tj]:
                if self.pIndex[t] not in tMap:
                    tMap[self.pIndex[t]] = 0
                tMap[self.pIndex[t]] += 1

            for k,v in tMap.items():
                pairInstSum += v*(v-1)/2.0

        try:
            pP =  pairInstSum / pairPrSum
        except ZeroDivisionError:
            pP = 1.0

        try:
            pR = pairInstSum / pairTrSum 
        except ZeroDivisionError:
            pR = 1.0

        return pP, pR, f1(pP, pR)




'''
Created on 2021-04-29

@author: lucianovilas@gmail.com

'''



import numpy as np



def make_clusters(labels_true,labels_pred):
    
    n_samples = len(labels_true)
    
    true_clusters = {}  # true cluster_id => set of sample indices
    pred_clusters = {}  # pred cluster_id => set of sample indices

    for i in range(n_samples):
        true_cluster_id = labels_true[i]
        pred_cluster_id = labels_pred[i]

        if true_cluster_id not in true_clusters:
            true_clusters[true_cluster_id] = set()
        if pred_cluster_id not in pred_clusters:
            pred_clusters[pred_cluster_id] = set()

        true_clusters[true_cluster_id].add(i)
        pred_clusters[pred_cluster_id].add(i)

    for cluster_id, cluster in true_clusters.items():
        true_clusters[cluster_id] = frozenset(cluster)
    for cluster_id, cluster in pred_clusters.items():
        pred_clusters[cluster_id] = frozenset(cluster)    


    return true_clusters, pred_clusters



def check_labels(labels_true, labels_pred):
    """Check that the two clusterings matching 1D integer arrays"""
    labels_true = np.asarray(labels_true)
    labels_pred = np.asarray(labels_pred)

    # input checks
    if labels_true.ndim != 1:
        raise ValueError(
            "labels_true must be 1D: shape is %r" % (labels_true.shape,))
    if labels_pred.ndim != 1:
        raise ValueError(
            "labels_pred must be 1D: shape is %r" % (labels_pred.shape,))
    if labels_true.shape != labels_pred.shape:
        raise ValueError(
            "labels_true and labels_pred must have same size, got %d and %d"
            % (labels_true.shape[0], labels_pred.shape[0]))
    return labels_true, labels_pred


def check_clusters(labels_true_c, labels_pred_c):
    """ Check that labels_true_c and labels_pred_c have the same number of instances """
    
    if len(labels_true_c) == 0:
            raise ValueError("labels_true_c must have at least one instance")
    if len(labels_pred_c) == 0:
            raise ValueError("labels_pred_c must have at least one instance")    

    l1, l2 = 0, 0
    for k in labels_true_c: l1 += len(labels_true_c[k]) 
    for k in labels_pred_c: l2 += len(labels_pred_c[k]) 

    if l1 != l2:
        raise ValueError('Cluster labels are not the same number of instances')
    
    return labels_true_c, labels_pred_c




import setuptools

with open('README.md', 'rb') as f:
    long_description = f.read().decode('utf-8')

setuptools.setup(
    name="nd_metrics",
    version="0.0.4",
    author="Luciano Vilas Boas Espiridião",
    author_email="lucianovilasboas@gmail.com",
    description="Metrics for Author Name Disambiguation (AND) evaluation methods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lucianovilasboas/nd_metrics",
    packages=["nd_metrics"],#setuptools.find_packages(),
    install_requires=['numpy'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license = 'MIT',
    keywords = 'name disambiguation metrics',    
    python_requires='>=3.6',
)